"""This module provides named electrical models.

Copyright 2021--2022 Michael Hayes, UCECE"""

from lcapy import R, L, C, Par, CPE
from .model import Model, modelmake

SeriesRLModel = modelmake('SeriesRLModel', R('R1') + L('L1'), ('R1', 'L1'))
SeriesRCModel = modelmake('SeriesRCModel', R('R1') + C('C1'), ('R1', 'C1'))
SeriesLCModel = modelmake('SeriesLCModel', L('L1') + C('C1'), ('L1', 'C1'))

ParallelRLModel = modelmake('ParallelRLModel', R('R1') | L('L1'), ('R1', 'L1'))
ParallelRCModel = modelmake('ParallelRCModel', R('R1') | C('C1'), ('R1', 'C1'))
ParallelLCModel = modelmake('ParallelLCModel', L('L1') | C('C1'), ('L1', 'C1'))

SeriesRLCModel = modelmake('SeriesRLCModel',
                           R('R1') + L('L1') + C('C1'), ('R1', 'L1', 'C1'))
ParallelRLCModel = modelmake('ParallelRLCModel',
                             Par(R('R1'),  L('L1'), C('C1')), ('R1', 'L1', 'C1'))

ParallelRSeriesRLModel = modelmake('ParallelRSeriesRLModel',
                                   R('R1') | (R('R2') + L('L1')), ('R1', 'R2', 'L1'))
ParallelCSeriesRLModel = modelmake('ParallelCSeriesRLModel',
                                   C('C1') | (R('R1') + L('L1')), ('C1', 'R1', 'L1'))
ParallelLSeriesRLModel = modelmake('ParallelLSeriesRLModel',
                                   L('L1') | (R('R1') + L('L2')), ('L1', 'R1', 'L2'))

ParallelRSeriesLCModel = modelmake('ParallelRSeriesLCModel',
                                   R('R1') | (L('L1') + C('C1')), ('R1', 'L1', 'C1'))
ParallelCSeriesLCModel = modelmake('ParallelCSeriesLCModel',
                                   C('C1') | (C('C2') + L('L1')), ('C1', 'C2', 'L1'))
ParallelLSeriesLCModel = modelmake('ParallelLSeriesLCModel',
                                   L('L1') | (C('C1') + L('L2')), ('L1', 'C1', 'L2'))

ParallelRSeriesRCModel = modelmake('ParallelRSeriesRCModel',
                                   R('R1') | (R('R1') + C('C1')), ('R1', 'R2', 'C1'))
ParallelCSeriesRCModel = modelmake('ParallelCSeriesRCModel',
                                   C('C1') | (R('R1') + C('C2')), ('C1', 'R1', 'C2'))
ParallelLSeriesRCModel = modelmake('ParallelLSeriesRCModel',
                                   L('L1') | (R('R1') + C('C1')), ('L1', 'R1', 'C1'))

SeriesRParallelRLModel = modelmake('SeriesRParallelRLModel',
                                   R('R1') + (R('R2') | L('L1')), ('R1', 'R2', 'L1'))
SeriesCParallelRLModel = modelmake('SeriesCParallelRLModel',
                                   C('C1') + (R('R1') | L('L1')), ('C1', 'R1', 'L1'))
SeriesLParallelRLModel = modelmake('SeriesLParallelRLModel',
                                   L('L1') + (R('R1') | L('L2')), ('L1', 'R1', 'L2'))

SeriesRParallelLCModel = modelmake('SeriesRParallelLCModel',
                                   R('R1') + (L('L1') | C('C1')),
                                   ('R1', 'L1', 'C1'))
SeriesCParallelLCModel = modelmake('SeriesCParallelLCModel',
                                   C('C1') + (C('C2') | L('L1')),
                                   ('C1', 'C2', 'L1'))
SeriesLParallelLCModel = modelmake('SeriesLParallelLCModel',
                                   L('L1') + (C('C1') | L('L2')),
                                   ('L1', 'C1', 'L2'))

SeriesRParallelRCModel = modelmake('SeriesRParallelRCModel',
                                   R('R1') + (R('R1') | C('C1')),
                                   ('R1', 'R2', 'C1'))
SeriesCParallelRCModel = modelmake('SeriesCParallelRCModel',
                                   C('C1') + (R('R1') | C('C2')),
                                   ('C1', 'R1', 'C2'))
SeriesLParallelRCModel = modelmake('SeriesLParallelRCModel',
                                   L('L1') + (R('R1') | C('C1')),
                                   ('L1', 'R1', 'C1'))

Parallel2SeriesRLModel = modelmake('Parallel2SeriesRLModel',
                                   (R('R1') + L('L1')) | (R('R2') + L('L2')),
                                   ('R1', 'L1', 'R2', 'L2'))
Parallel3SeriesRLModel = modelmake('Parallel3SeriesRLModel',
                                   (R('R1') + L('L1')) | (R('R2') + L('L2')) |
                                   (R('R3') + L('L3')),
                                   ('R1', 'L1', 'R2', 'L2', 'R3', 'L3'))
Parallel4SeriesRLModel = modelmake('Parallel4SeriesRLModel',
                                   (R('R1') + L('L1')) | (R('R2') + L('L2')) |
                                   (R('R3') + L('L3')) | (R('R4') + L('L4')),
                                   ('R1', 'L1', 'R2', 'L2', 'R3', 'L3', 'R4', 'L4'))

CPEModel = modelmake('CPEModel', CPE('K', 'alpha'), ('K', 'alpha'))
SeriesRCPEModel = modelmake('SeriesRCPEModel', R(
    'R1') + CPE('K', 'alpha'), ('R1', 'K', 'alpha'))
ParallelRCPEModel = modelmake('ParallelRCPEModel', R(
    'R1') | CPE('K', 'alpha'), ('R1', 'K', 'alpha'))

SeriesRParallelRCPEModel = modelmake('SeriesRParallelRCPEModel',
                                     R('R1') + (R('R2') | CPE('K', 'alpha')),
                                     ('R1', 'R2', 'K', 'alpha'))
ParallelRSeriesRCPEModel = modelmake('ParallelRSeriesRCPEModel',
                                     R('R1') | (R('R2') + CPE('K', 'alpha')),
                                     ('R1', 'R2', 'K', 'alpha'))



"""This module parses impedance data.

Copyright 2021--2025 Michael Hayes, UCECE"""

from numpy import loadtxt, pi, sin, cos, radians, exp
from os.path import basename, splitext


class ImpedanceData(object):

    def __init__(self, filename, f, Z, fmin=None, fmax=None, conjugate=False):

        if fmin is not None:
            m = f >= fmin
            f = f[m]
            Z = Z[m]
        if fmax is not None:
            m = f <= fmax
            f = f[m]
            Z = Z[m]

        if conjugate:
            Z = Z.conjugate()

        self.f = f
        self.Z = Z
        self.filename = basename(filename)
        self.name, ext = splitext(self.filename)

    @property
    def Y(self):
        return 1 / self.Z


class KeysightE4990AImpedanceData(ImpedanceData):

    def __init__(self, filename, fmin, fmax, conjugate):

        lines = open(filename).readlines()
        if not lines[0].startswith('!Agilent Technologies,E4990A'):
            raise ValueError('Not Keysight E4990A')

        if lines[4].startswith('Frequency(Hz), |Z|(Ohm)-data, theta-z(deg)-data'):
            foo = loadtxt(filename, skiprows=5,
                          delimiter=',', comments='END')
            f = foo[:, 0]
            A = foo[:, 1]
            theta = radians(foo[:, 2])
            Z = (cos(theta) + 1j * sin(theta)) * A

        elif lines[4].startswith('Frequency(Hz), R(Ohm)-data, X(Ohm)-data'):

            foo = loadtxt(filename, skiprows=5,
                          delimiter=',', comments='END')
            f = foo[:, 0]
            Z = foo[:, 1] + 1j * foo[:, 2]

        elif lines[4].startswith('Frequency(Hz), Ls(H)-data, Rs(Ohm)-data'):

            foo = loadtxt(filename, skiprows=5,
                          delimiter=',', comments='END')
            f = foo[:, 0]
            Ls = foo[:, 1]
            Rs = foo[:, 2]
            Z = Rs + 1j * 2 * pi * f * Ls
        elif lines[4].startswith('Frequency(Hz), |Z|(Ohm)-data'):
            raise ValueError('No phase information for Keysight E4990A')
        else:
            raise ValueError('Unhandled format for Keysight E4990A')

        super(KeysightE4990AImpedanceData, self).__init__(filename, f,
                                                          Z, fmin, fmax,
                                                          conjugate)


class GenericImpedanceData(ImpedanceData):

    def __init__(self, filename, fmin, fmax, conjugate,
                 admittance=False, magphase=False):

        foo = loadtxt(filename, skiprows=0, delimiter=',', comments='#')
        if foo.shape[1] != 3:
            raise ValueError('Expecting 3 columns (frequency, real, imag')

        f = foo[:, 0]

        if magphase:
            Z = foo[:, 1] * exp(1j * radians(foo[:, 2]))
        else:
            Z = foo[:, 1] + 1j * foo[:, 2]

        if admittance:
            Z = 1 / Z

        super(GenericImpedanceData, self).__init__(filename, f, Z,
                                                   fmin, fmax, conjugate)


def impedancedata(filename, fmin=None, fmax=None, conjugate=False,
                  admittance=False, magphase=False):

    try:
        open(filename)
    except:
        raise ValueError('Cannot read file %s' % filename)

    try:
        return KeysightE4990AImpedanceData(filename, fmin, fmax, conjugate)
    except:
        pass

    try:
        return GenericImpedanceData(filename, fmin, fmax, conjugate,
                                    admittance, magphase)
    except:
        pass

    # Add support for other data formats.

    raise ValueError('Cannot determine impedance format for %s' % filename)



#!/usr/bin/python3
"""zfitpy V0.3.7
Copyright (c) 2021--2025 Michael P. Hayes, UC ECE, NZ

Usage:

Here are some examples.

zfitpy --net net --draw
zfitpy --input data.csv --plot --output plotfilename
zfitpy --plot-fit --net net --ranges ranges --input data.csv --plot-error
zfitpy --net net --laplace
"""

from __future__ import print_function
from matplotlib.pyplot import show, savefig, style
from argparse import ArgumentParser
from zfitpy.model import models, modelmake
from zfitpy import ZFitter
from zfitpy import Plotter
from zfitpy import impedancedata
import sys


def zfitpy_exception(type, value, tb):
    if hasattr(sys, 'ps1') or not sys.stderr.isatty():
        # We are not in interactive mode or we don't have a tty-like
        # device, so call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import traceback
        import pdb
        # We are in interactive mode, print the exception...
        traceback.print_exception(type, value, tb)
        print()
        # ...then start the debugger in post-mortem mode.
        pdb.pm()


def model_make(args):

    if not args.net and not args.modelname:
        raise ValueError(
            'Either need to specify model name with --modelname or network with --net')

    if args.net:
        net = args.net
        if net.endswith('.net'):
            net = open(net).read()

        Model = modelmake('Model', net)

    if args.modelname:
        try:
            Model = models[args.modelname]
        except:
            modelnames = ', '.join(list(models.keys()))
            raise ValueError('Unknown model %s: known models: %s' %
                             (args.modelname, modelnames))

    return Model


def main():

    parser = ArgumentParser(description='Draw schematic of impedance model.')
    parser.add_argument('--version', action='version',
                        version=__doc__.split('\n')[0])
    parser.add_argument('--modelname', type=str, help='model name')
    parser.add_argument('--net', type=str,
                        help="specify network, e.g., R('R1') + L('L1')")
    parser.add_argument('--input_filename', type=str, help='input filename')
    parser.add_argument('--output_filename', type=str, help='output filename')
    parser.add_argument(
        '--ranges', type=str, help="specify search ranges, e.g.,  {'R1':(0,1),'L1':(10,20)}")
    parser.add_argument('--draw', action='store_true',
                        default=False, help='draw network')
    parser.add_argument('--layout', type=str, default='horizontal',
                        help='drawing layout: vertical, horizontal, ladder')
    parser.add_argument('--show', action='store_true',
                        default=False, help='show plot')
    parser.add_argument('--nyquist', action='store_true',
                        default=False, help='use Nyquist plot')
    parser.add_argument('--admittance', action='store_true',
                        default=False,
                        help='plot admittance rather than impedance')
    parser.add_argument('--logf', action='store_true',
                        default=False, help='Plot log frequency')
    parser.add_argument('--plot-error', action='store_true',
                        default=False, help='plot impedance error')
    parser.add_argument('--plot-fit', action='store_true',
                        default=False, help='plot impedance data and fit')
    parser.add_argument('--plot-data', action='store_true',
                        default=False, help='plot impedance data')
    parser.add_argument('--magphase', action='store_true',
                        default=False, help='plot magnitude and phase of impedance')
    parser.add_argument('--title', type=str, help='title for plot')
    parser.add_argument(
        '--method', type=str, help='optimization method: brute, trf, or dogbox', default='brute')
    parser.add_argument('--steps', type=int, default=20,
                        help='the number of search steps per range')
    parser.add_argument('--fmin', type=float, default=None,
                        help='minimum frequency to use for fitting')
    parser.add_argument('--fmax', type=float, default=None,
                        help='maximum frequency to use for fitting')
    parser.add_argument('--finish', type=str,
                        help='finishing search method: none or fmin')
    parser.add_argument('--verbose', type=int, default=0,
                        help='set verbosity 0-2')
    parser.add_argument('--conjugate', action='store_true',
                        help='conjugate the impedance')
    parser.add_argument('--pdb', action='store_true', default=False,
                        help='enter python debugger on exception')
    parser.add_argument('--defs', action='store_true', default=False,
                        help='print component definitions as a dictionary')
    parser.add_argument('--error', action='store_true', default=False,
                        help='print the fitting error')
    parser.add_argument('--values', action='store_true', default=False,
                        help='print the fitted values')
    parser.add_argument('--style', type=str, help='matplotlib style filename')
    parser.add_argument('--laplace', action='store_true', default=False,
                        help='print impedance in Laplace domain')
    parser.add_argument('--Ls', action='store_true', default=False,
                        help='plot series inductance')
    parser.add_argument('--Rs', action='store_true', default=False,
                        help='plot series resistance')
    parser.add_argument('--Zoffset', type=float, default=0,
                        help='impedance offset to add')
    parser.add_argument('--sigfigs', type=int, default=10,
                        help='set number of significant figures when printing defs')
    parser.add_argument('--data-admittance', action='store_true',
                        default=False,
                        help='interpret data as admittance')
    parser.add_argument('--data-magphase', action='store_true',
                        default=False,
                        help='interpret data as magnitude/phase')
    parser.add_argument('--percent', action='store_true',
                        default=False,
                        help='show fitting error as percentage')

    args = parser.parse_args()

    if args.pdb:
        sys.excepthook = zfitpy_exception

    if args.style:
        style.use(args.style)

    if args.laplace:
        Model = model_make(args)
        model = Model()
        print(model.net.Z.laplace())
        return 0

    if args.draw:
        Model = model_make(args)
        model = Model()
        model.draw(args.output_filename, layout=args.layout)

        if args.show or args.output_filename is None:
            show()
        return 0

    if args.input_filename is None:
        raise ValueError('Impedance data not specified')

    data = impedancedata(args.input_filename,
                         fmin=args.fmin, fmax=args.fmax,
                         conjugate=args.conjugate,
                         admittance=args.data_admittance,
                         magphase=args.data_magphase)
    if args.Zoffset != 0:
        data.Z += args.Zoffset

    if False and args.ranges is None:
        raise ValueError('Search ranges not specified')

    if args.ranges is None:
        fitmodel = None
    else:

        ranges = args.ranges
        if ranges.endswith('.ranges'):
            ranges = open(ranges).read()

        Model = model_make(args)
        zfitter = ZFitter(Model, data.f, data.Z)
        fitmodel = zfitter(ranges=ranges, Ns=args.steps,
                           method=args.method, verbose=args.verbose,
                           finish=args.finish)
        if args.error:
            print('error=%.3e' % fitmodel.error)
        if args.defs:
            print(fitmodel.defs(args.sigfigs))
        if args.values:
            print(fitmodel)
        if not (args.error or args.defs or args.values):
            print('%s, error=%.3e' % (fitmodel, fitmodel.error))

    plotter = Plotter(args.admittance, args.logf)
    if args.plot_error and fitmodel:
        plotter.error(data, fitmodel, title=args.title, percent=args.percent)

    if args.plot_fit:
        if args.nyquist:
            plotter.nyquist(data, fitmodel, title=args.title)
        elif args.Ls or args.Rs:
            plotter.LsRs_fit(data, fitmodel, title=args.title,
                             doLs=args.Ls, doRs=args.Rs)
        else:
            plotter.fit(data, fitmodel, title=args.title,
                        magphase=args.magphase)

    if args.plot_data:
        if args.nyquist:
            plotter.nyquist(data, None, title=args.title)
        elif args.Ls or args.Rs:
            plotter.LsRs_fit(data, None, title=args.title,
                             doLs=args.Ls, doRs=args.Rs)
        else:
            plotter.data(data, title=args.title, magphase=args.magphase)

    if args.output_filename is not None:
        savefig(args.output_filename, bbox_inches='tight')

    if args.show or args.output_filename is None:
        show()

    return 0


if __name__ == '__main__':
    main()



"""Zfitpy is a Python library for fitting electrical one-port models
to measured impedance data.  The electrical network is specified using
Lcapy notation, such as R('R1') + L('L1').

Copyright 2021 Michael Hayes, UCECE

"""
name = "zfitpy"

import pkg_resources

from .zfitter import *
from .zfitterbrute import *
from .zfittercurve import *
from .model import *
from .plotter import *
from .impedancedata import *
from .models import *
from .zfit import *

__version__ = pkg_resources.require('zfitpy')[0].version



"""This module is a wrapper for the SciPy optimizers

Copyright 2021 Michael Hayes, UCECE"""

from scipy.optimize import brute, fmin
from .zfitterbase import ZFitterBase


class ZFitterBrute(ZFitterBase):
    """Class for non-linear least-squares using the brute force
    SciPy algorithm."""

    def Zerror_params(self, params):

        model = self._model(*params)
        rmse = model.Zrmse(self.f, self.Z)
        if self.verbose > 1:
            print(model, rmse)
        return rmse

    def Yerror_params(self, params):

        model = self._model(*params)
        rmse = model.Yrmse(self.f, self.Y)
        if self.verbose > 1:
            print(model, rmse)
        return rmse

    def optimize(self, ranges=None, opt='Z', Ns=10, finish='fmin', **kwargs):
        """Ranges is a list of tuples, of the form: (min, max) or (min, max,
        numsteps).  If `numsteps` is not specified then `Ns` is used."""

        kwargs.pop('method', None)
        self.verbose = kwargs.pop('verbose', 0)

        if finish in ('none', 'None', ''):
            finish = None
        elif finish == 'fmin':
            finish = fmin

        if opt == 'Z':
            func = self.Zerror_params
        elif opt == 'Y':
            func = self.Yerror_params
        else:
            raise ValueError("Opt must be 'Y' or 'Z'")

        ranges = self._make_ranges(ranges)

        oranges = []
        for r in ranges:
            if len(r) == 2:
                # Note, a complex value specifies the number of steps
                # (see numpy.mgrid).
                oranges.append(slice(r[0], r[1], complex(Ns)))
            elif len(r) == 3:
                oranges.append(slice(r[0], r[1], complex(r[2])))
            else:
                raise ValueError('Range %s can only have 2 or 3 values' % r)

        params, rmse, foo, bar = brute(func, oranges, Ns=Ns, args=(),
                                       finish=finish, full_output=1)

        model = self._model(*params)
        return model, rmse, None



"""This module provides plotting support for impedance data.

Copyright 2021 Michael Hayes, UCECE"""

from matplotlib.pyplot import subplots
from numpy import degrees, angle, pi


class Plotter:

    def __init__(self, admittance=False, logf=False):

        self.admittance = admittance
        self.logf = logf

        if self.admittance:
            self.label = 'Admittance'
            self.units = 'siemens'
        else:
            self.label = 'Impedance'
            self.units = 'ohms'

    def set_title(self, axes, title=None, model=None, sfmax=2):

        if title is None:
            if model is None:
                return axes
            title = str(model)

        if model is not None:
            if r'%rmse' in title:
                title = title.replace(r'%rmse', '%.2e' % model._rmse)
            if r'%method' in title:
                title = title.replace(r'%method', str(model._method))
            if r'%model' in title:
                title = title.replace(r'%model', model.title(sfmax=sfmax))

        axes.set_title(title)
        return axes

    def error(self, data, model, axes=None, title=None, percent=False):

        mZ = model.Z(data.f)
        dZ = data.Z
        if self.admittance:
            mZ = 1 / mZ
            dZ = 1 / dZ

        if axes is None:
            fig, axes = subplots(1)

        if self.logf:
            plot = axes.semilogx
        else:
            plot = axes.plot

        ereal = dZ.real - mZ.real
        eimag = dZ.imag - mZ.imag
        if percent:
            ereal = ereal / dZ.real * 100
            eimag = eimag / dZ.imag * 100

        plot(data.f, ereal, label=data.name + ' real')
        plot(data.f, eimag, '--', label=data.name + ' imag')

        axes.set_xlabel('Frequency (Hz)')
        if percent:
            axes.set_ylabel('Percent error')
        else:
            axes.set_ylabel(f'{self.label} error ({self.units})')
        axes.grid(True)

        self.set_title(axes, title, model)
        axes.legend()
        return axes

    def data(self, data, axes=None, title=None, magphase=False):

        self.fit(data, None, axes, title, magphase)

    def fit(self, data, model=None, axes=None, title=None, magphase=False):

        if model is not None:
            mZ = model.Z(data.f)
            if self.admittance:
                mZ = 1 / mZ
        else:
            mZ = None

        dZ = data.Z
        if self.admittance:
            dZ = 1 / dZ

        if axes is None:
            fig, axes = subplots(1)

        if self.logf:
            plot = axes.semilogx
        else:
            plot = axes.plot

        if magphase:
            axes2 = axes.twinx()

            if self.logf:
                plot2 = axes2.semilogx
            else:
                plot2 = axes2.plot

            plot(data.f, abs(dZ), label=data.name + ' data mag')
            plot2(data.f, degrees(angle(dZ)), '--',
                  label=data.name + ' data phase')
            if mZ is not None:
                plot(data.f, abs(mZ), label=data.name + ' model mag')
                plot2(data.f, degrees(angle(mZ)), '--',
                      label=data.name + ' model phase')
            axes2.legend()
            axes2.set_ylabel(f'{self.label} phase (degrees)')

        else:
            plot(data.f, dZ.real, label=data.name + ' data real')
            plot(data.f, dZ.imag, '--',
                 label=data.name + ' data imag')

            if mZ is not None:
                plot(data.f, mZ.real, label=data.name + ' model real')
                plot(data.f, mZ.imag, '--',
                     label=data.name + ' model imag')

        axes.set_xlabel('Frequency (Hz)')
        axes.set_ylabel(f'{self.label} ({self.units})')
        axes.grid(True)

        self.set_title(axes, title, model)
        axes.legend()
        return axes

    def nyquist(self, data, model=None, axes=None, title=None):

        dZ = data.Z
        if self.admittance:
            dZ = 1 / dZ

        if model is not None:
            mZ = model.Z(data.f)
            if self.admittance:
                mZ = 1 / mZ
        else:
            mZ = None

        if axes is None:
            fig, axes = subplots(1)

        axes.plot(dZ.real, dZ.imag, label=data.name + ' data')
        if mZ is not None:
            axes.plot(mZ.real, mZ.imag, label=data.name + ' model')

        axes.set_xlabel(f'{self.label} real ({self.units})')
        axes.set_ylabel(f'{self.label} imag ({self.units})')
        axes.grid(True)

        self.set_title(axes, title, model)
        axes.legend()
        return axes

    def difference(self, data1, data2, axes=None, title=None):

        dZ1 = data1.Z
        dZ2 = data2.Z
        if self.admittance:
            dZ1 = 1 / dZ1
            dZ2 = 1 / dZ2

        if axes is None:
            fig, axes = subplots(1)

        if not (data1.f == data2.f).all():
            raise ValueError('Mismatched frequencies')

        Zdiff = dZ1 - dZ2

        name = '(%s - %s)' % (data1.name, data2.name)

        axes.plot(data1.f, Zdiff.real, label=name + ' real')
        axes.plot(data1.f, Zdiff.imag, '--', label=name + ' imag')

        axes.set_xlabel('Frequency (Hz)')
        axes.set_ylabel(f'{self.label} error ({self.units})')
        axes.grid(True)

        self.set_title(axes, title, model=None)
        axes.legend()
        return axes

    def LsRs_fit(self, data, model=None, axes=None, title=None,
                 doLs=True, doRs=True):

        if model is not None:
            Z = model.Z(data.f)
        else:
            Z = None

        if axes is None:
            fig, axes = subplots(1)

        if doLs and doRs:
            Lsaxes = axes

            # Handle multiple calls with specified axes
            if not hasattr(axes, '_twinx'):
                axes._twinx = axes.twinx()

            Rsaxes = axes._twinx
        elif doLs:
            Lsaxes = axes
            Rsaxes = None
        else:
            Rsaxes = axes
            Lsaxes = None

        if Rsaxes is not None:
            Rs = data.Z.real
            Rsaxes.plot(data.f, Rs, '--', label=data.name + ' data Rs')
            Rsaxes.set_ylabel('Resistance (ohm)')
            if Lsaxes is not None:
                # Dummy plot to advance colour.
                # Lsaxes.plot([], [])
                pass

        if Z is not None:
            if Rsaxes is not None:
                Rs = Z.real
                Rsaxes.plot(data.f, Rs, '--', label=data.name + ' model Rs')
            if Lsaxes is not None:
                # Dummy plot to advance colour.
                # Lsaxes.plot([], [])
                pass

        if Lsaxes is not None:
            Ls = data.Z.imag / (2 * pi * data.f)
            Lsaxes.plot(data.f, Ls * 1e3, label=data.name + ' data Ls')
            Lsaxes.set_ylabel('Inductance (mH)')

        if Z is not None:
            if Lsaxes is not None:
                Ls = Z.imag / (2 * pi * data.f)
                Lsaxes.plot(data.f, Ls * 1e3, label=data.name + ' model Ls')

        axes.set_xlabel('Frequency (Hz)')
        axes.grid(True)

        self.set_title(axes, title, model)

        if Lsaxes is not None:
            Lsaxes.legend()
        if Rsaxes is not None:
            Rsaxes.legend()
        return axes



"""This module is a wrapper for the SciPy optimizers

Copyright 2021 Michael Hayes, UCECE"""

from .zfitterbrute import ZFitterBrute
from .zfittercurve import ZFitterCurve


class ZFitter(object):

    def __init__(self, model, f, Z):
        self._model = model
        self.f = f
        self.Z = Z

    def __call__(self, ranges=None, opt=None, method='brute', fmin=None, fmax=None,
                 **kwargs):

        parts = method.split('-')
        if len(parts) > 2:
            raise ValueError('Unknown method %s' % method)
        if len(parts) == 2:
            if parts[1] not in ('Y', 'Z'):
                raise ValueError('Unknown method objective %s' % parts[1])
            if opt is not None:
                raise ValueError('Cannot specify method objective and opt')
            opt = parts[1]
        else:
            if opt is None:
                opt = 'Z'
        method = parts[0]

        f = self.f
        Z = self.Z
        if fmin is not None:
            m = f >= fmin
            f = f[m]
            Z = Z[m]
        if fmax is not None:
            m = f <= fmax
            f = f[m]
            Z = Z[m]

        if method == 'brute':
            fitter = ZFitterBrute(self._model, f, Z)
        elif method in ('trf', 'dogbox'):
            fitter = ZFitterCurve(self._model, f, Z)
        else:
            raise ValueError(
                'Unknown method %s: needs to be brute, trf, dogbox' % method)

        model, rmse, cov = fitter.optimize(
            ranges, opt, method=method, **kwargs)
        model._rmse = rmse
        model._cov = cov
        model._method = method

        return model



"""This module provides the impedance Model class.

Models can be dynamically created given a network described using
Lcapy network syntax, e.g., "R('R1') + L('L1') | R('R2')".

Impedances are calculated using Lcapy.  To speed up evaluations, the
impedance calculating code is lazily compiled and cached.

Copyright 2021 Michael Hayes, UCECE

"""

from lcapy import f, t, G, R, C, L, CPE, Par, Ser, jf
from math import floor, log10
from numpy import mean, sqrt, pi as npi
from .engformatter import EngFormatter

models = {}


def round_to_n(x, n):
    """Round to n significant figures."""

    if x == 0:
        return x

    return round(x, -int(floor(log10(abs(x)))) + (n - 1))


def modelmake(name, net, paramnames=None):

    if isinstance(net, str):
        net = eval(net)

    if paramnames is None:
        try:
            # For Lcapy 0.99
            paramnames = net.params
        except:
            Z = eval('(%s).Z(jf)' % net)
            paramnames = Z.symbols
            paramnames.pop('t', None)
            paramnames.pop('f', None)

    params = ', '.join(paramnames)

    docstring = name + '(' + params + '): ' + str(net)
    newclass = type(name, (Model, ), {'__doc__': docstring})
    newclass._net = net
    newclass.paramnames = paramnames
    models[name] = newclass
    return newclass


class Model(object):

    _Ycode = None
    _Zcode = None
    _rmse = 0
    _cov = None
    _method = ''

    def __init__(self, *args):
        """Create model instance."""

        if len(args) == 0:
            return

        for m, p in enumerate(self.paramnames):
            setattr(self, p, args[m])

    def __repr__(self):
        return self.__str__()

    def __call__(self, i, t):

        return self.v(i, t)

    def v(self, i, t):
        """Calculate voltage drop across the network given a current signal."""

        return self._net.subs(vars(self)).Z.response(i, t)

    def i(self, v, t):
        """Calculate current through the network given an applied voltage
        signal."""

        return self._net.subs(vars(self)).Y.response(v, t)

    def draw(self, filename=None, layout='horizontal'):
        """Draw the network."""

        if filename is not None and filename.endswith('.sch'):
            outfile = open(filename, 'w')
            sch = self._net.sch(layout=layout)
            outfile.write(str(sch))
            outfile.close()
        else:
            self._net.draw(filename, layout=layout)

    def _build(self, foo, name):
        paramnames = foo.symbols
        paramnames.pop('t', None)
        paramnames.pop('f', None)

        # FIXME if have R1 and R since the replacement will fail.
        codestr = str(foo)
        for p in paramnames:
            codestr = codestr.replace(p, 'self.' + p)
        code = compile(codestr, name, 'eval')

        return code

    def _Ybuild(self):

        return self._build(self._net.Y(jf), 'Y')

    def _Zbuild(self):

        return self._build(self._net.Z(jf), 'Z')

    def Z(self, f):
        """Return impedance at frequency `f`; `f` can be an ndarray."""

        if self._Zcode is None:
            # Cache result for class
            self.__class__._Zcode = self._Zbuild()

        j = 1j
        pi = npi

        return eval(self._Zcode)

    def Y(self, f):
        """Return admittance at frequency `f`; `f` can be an ndarray."""

        if self._Ycode is None:
            # Cache result for class
            self.__class__._Ycode = self._Ybuild()

        j = 1j
        pi = npi

        return eval(self._Ycode)

    def __str__(self):

        return self.title(sfmax=3)

    def title(self, sfmax=3):

        parts = []
        for var, val in vars(self).items():
            if var[0] == '_':
                continue

            try:
                units = {'R': 'ohm', 'C': 'F',
                         'L': 'H', 'K': '', 'a': ''}[var[0]]
            except KeyError:
                units = ''

            parts.append('%s=%s' %
                         (var, EngFormatter(hundreds=True, sfmax=sfmax).str(val, units)))

        return ', '.join(parts)

    def defs(self, dp=10):

        defs = {}
        for var, val in vars(self).items():
            if var[0] == '_':
                continue
            defs[var] = round_to_n(val, dp)

        return defs

    @property
    def net(self):
        return self._net.subs(vars(self))

    def Yrmse(self, f, Y):

        Yerr = Y - self.Y(f)
        rmse = sqrt(mean(Yerr.real**2 + Yerr.imag**2))
        return rmse

    def Zrmse(self, f, Z):

        Zerr = Z - self.Z(f)
        rmse = sqrt(mean(Zerr.real**2 + Zerr.imag**2))
        return rmse

    @property
    def error(self):
        return self._rmse



"""This module is a wrapper for the SciPy optimizers

Copyright 2021 Michael Hayes, UCECE"""

import numpy as np
from scipy.optimize import curve_fit
from .zfitterbase import ZFitterBase


class ZFitterCurve(ZFitterBase):
    """Class for non-linear least-squares using the trf and dogbox
    SciPy algorithms."""

    def Z_params(self, f, *params):
        Z = self._model(*params).Z(f)
        return np.hstack((Z.real, Z.imag))

    def Y_params(self, f, *params):
        Y = self._model(*params).Y(f)
        return np.hstack((Y.real, Y.imag))

    def optimize(self, ranges=None, opt='Z', ftol=1e-14, xtol=1e-14,
                 maxfev=1e5, **kwargs):
        """Ranges is a list of tuples, of the form: (min, max)."""

        kwargs.pop('Ns', None)
        kwargs.pop('finish', None)

        if opt == 'Z':
            func = self.Z_params
            ydata = self.Z
        elif opt == 'Y':
            func = self.Y_params
            ydata = self.Y
        else:
            raise ValueError("Opt must be 'Z' or 'Y'")

        ranges = self._make_ranges(ranges)

        bounds_min = np.zeros(len(ranges))
        bounds_max = np.zeros(len(ranges))

        for m, r in enumerate(ranges):
            if len(r) in (2, 3):
                bounds_min[m] = r[0]
                bounds_max[m] = r[1]
            else:
                raise ValueError('Range %s can only have 2 or 3 values' % r)

        bounds = (bounds_min, bounds_max)

        # Initial guess.
        p0 = 0.5 * (bounds_min + bounds_max)

        ydata = np.hstack((ydata.real, ydata.imag))
        params, cov = curve_fit(func, self.f, ydata, p0=p0,
                                bounds=bounds, ftol=ftol, xtol=xtol,
                                maxfev=maxfev, **kwargs)
        #err = np.sqrt(np.diag(cov))

        model = self._model(*params)

        rmse = model.Zrmse(self.f, self.Z)
        return model, rmse, cov



#!/usr/bin/env python
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

tests_require = ['nose', 'flake8', 'flake8-bugbear', 'flake8-comprehensions', 'flake8-requirements']

setup(name='zfitpy',
      version='0.5.1',
      author='Michael Hayes',
      author_email='michael.hayes@canterbury.ac.nz',
      description='Electrical model fitting to impedance data',
      long_description = long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/mph-/zfitpy',
      download_url='https://github.com/mph-/zfitpy',
      install_requires=['matplotlib',
                        'numpy',
                        'scipy',
                        'lcapy',
                        'setuptools',
      ],
      python_requires='>=3.6',
      extras_require={
          'test': tests_require,
          'doc': ['sphinx', 'ipython'],
          'release': ['wheel', 'twine'],
      },
      tests_require=tests_require,
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'zfitpy=zfitpy.scripts.zfitpy:main',
          ],
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
          "Operating System :: OS Independent",
      ],
)



"""This module provides a simple function for fitting an electric model to measured impedance data.

Copyright 2021 Michael Hayes, UCECE"""

from .impedancedata import ImpedanceData, impedancedata
from .model import modelmake, Model
from .zfitter import ZFitter


def zfit(data, model, ranges, method='brute', opt='Z', **kwargs):
    """Fit impedance data to a model.

    `data` can be a filename string or and `ImpedanceData` object
    `model` can be a string describing the network model or a `Model` class
    `ranges` is a dictionary of the search ranges for each component
    `model` can be 'brute', 'trf', or 'dogbox'
    `opt` can be `Y` for admittance fitting or `Z` for impedance fitting

    Returns: `ImpedanceData` object and a best-fit `Model` object.
    """

    if not isinstance(data, ImpedanceData):
        data = impedancedata(data)

    if not isinstance(model, Model):
        model = modelmake('Model', model)

    zfitter = ZFitter(model, data.f, data.Z)

    fitmodel = zfitter(ranges=ranges, method=method, opt=opt, **kwargs)

    return data, fitmodel



"""This module is a wrapper for the SciPy optimizers

Copyright 2021 Michael Hayes, UCECE"""


class ZFitterBase(object):

    def __init__(self, model, f, Z):
        self._model = model
        self.verbose = 0
        self.f = f
        self.Z = Z

    @property
    def Y(self):
        return 1 / self.Z

    def _make_ranges(self, ranges):

        if isinstance(ranges, str):
            ranges = eval(ranges)

        if isinstance(ranges, dict):
            rangesdict = ranges
            ranges = []
            for paramname in self._model.paramnames:
                ranges.append(rangesdict[paramname])

        return ranges
